#!usr/bin/perl

#############################################################################
#                                                                           #
# TCS230Compass.pl Perl/Tk program for calibrating a "tuna can" compass     #
#                  based on the TCS230 Color Sensor Module Set. It requires #
#                  the Tk module library, the BASIC Stamp program,          #
#                  TCS230Monitor.bs2, and an available serial port.         #
#                  This program is designed to run under both Windows and   #
#                  Linux.                                                   #
#                                                                           #
# Copyright (C) 2003  Bueno Systems, Inc.                                   #
#                                                                           #
# Contact: Phil Pilgrim: author@buenosystems.com                            #
#                                                                           #
# This program is free software; you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by      #
# the Free Software Foundation; version 2 of the License.                   #
#                                                                           #
# This program is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of            #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             #
# GNU General Public License for more details.                              #
#                                                                           #
# You should have received a copy of the GNU General Public License         #
# along with this program; if not, write to the Free Software               #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA #
#                                                                           #
#############################################################################

use strict;
my $NewSerialPort;
my @Devices;
my $DirDelim;

#Uncomment the following if compiling with perl2exe for Windows.

#use Win32::API;
#use Win32::SerialPort;

BEGIN {
	my $OpSys = lc $^O;
	if ($OpSys =~ m/linux/) {
		eval 'use Device::SerialPort qw(:STAT)';
		$NewSerialPort = sub{Device::SerialPort->new(@_)};
		@Devices = qw(/dev/ttyS0 /dev/ttyS1 /dev/ttyS2 /dev/ttyS3);
		$DirDelim = '/'
	} elsif ($OpSys =~ m/win/) {
		eval 'use Win32::SerialPort qw(:STAT)';
		$NewSerialPort = sub{Win32::SerialPort->new(@_)};
		@Devices = qw(COM1 COM2 COM3 COM4 COM5 COM6 COM7 COM8);
		$DirDelim = '\\'
	} else {
		die "Unknown operating System."
	}
}

use Carp;
use Tk;
use Tk::Menu;
use Tk::Canvas;
use Tk::Scale;
use Tk::DialogBox;
use Tk::ROText;
use Tk::Photo;
use Time::HiRes qw(time sleep);
use File::Basename;
use Cwd;
use GD;

my @RGB = qw(#ff0000 #00ff00 #0000ff);
my (@CurrentRGB, @CurrentHSV, @DefinedHue, @HueMap, @Periods);
my ($stream, $lastRGB, $Selected, $PrevSelected, $Flash, $AllDefined);
my $Busy = 0;
(my $DefaultDir = cwd()) =~ s:/:$DirDelim:g;
undef $/;
my $imgHSV = GetColorWheel();

my $imgLeft = <<END_OF_DATA;
R0lGODlhCAANAPcAAAAAAAEBAQICAgMDAwQEBAUFBQYGBgcHBwgICAkJCQoKCgsLCwwMDA0NDQ4O
Dg8PDxAQEBERERISEhMTExQUFBUVFRYWFhcXFxgYGBkZGRoaGhsbGxwcHB0dHR4eHh8fHyAgICEh
ISIiIiMjIyQkJCUlJSYmJicnJygoKCkpKSoqKisrKywsLC0tLS4uLi8vLzAwMDExMTIyMjMzMzQ0
NDU1NTY2Njc3Nzg4ODk5OTo6Ojs7Ozw8PD09PT4+Pj8/P0BAQEFBQUJCQkNDQ0REREVFRUZGRkdH
R0hISElJSUpKSktLS0xMTE1NTU5OTk9PT1BQUFFRUVJSUlNTU1RUVFVVVVZWVldXV1hYWFlZWVpa
WltbW1xcXF1dXV5eXl9fX2BgYGFhYWJiYmNjY2RkZGVlZWZmZmdnZ2hoaGlpaWpqamtra2xsbG1t
bW5ubm9vb3BwcHFxcXJycnNzc3R0dHV1dXZ2dnd3d3h4eHl5eXp6ent7e3x8fH19fX5+fn9/f4CA
gIGBgYKCgoODg4SEhIWFhYaGhoeHh4iIiImJiYqKiouLi4yMjI2NjY6Ojo+Pj5CQkJGRkZKSkpOT
k5SUlJWVlZaWlpeXl5iYmJmZmZqampubm5ycnJ2dnZ6enp+fn6CgoKGhoaKioqOjo6SkpKWlpaam
pqenp6ioqKmpqaqqqqurq6ysrK2tra6urq+vr7CwsLGxsbKysrOzs7S0tLW1tba2tre3t7i4uLm5
ubq6uru7u7y8vL29vb6+vr+/v8DAwMHBwcLCwsPDw8TExMXFxcbGxsfHx8jIyMnJycrKysvLy8zM
zM3Nzc7Ozs/Pz9DQ0NHR0dLS0tPT09TU1NXV1dbW1tfX19jY2NnZ2dra2tvb29zc3N3d3d7e3t/f
3+Dg4OHh4eLi4uPj4+Tk5OXl5ebm5ufn5+jo6Onp6erq6uvr6+zs7O3t7e7u7u/v7/Dw8PHx8fLy
8vPz8/T09PX19fb29vf39/j4+Pn5+fr6+vv7+/z8/P39/f7+/v///ywAAAAACAANAAAILQCZCRwI
YCBBZv8MFkR4UGBCZgsdQjSIMOLAhGgoPsx4cSBHhh4lGsz4cOS/gAA7
END_OF_DATA

my $imgRight = <<END_OF_DATA;
R0lGODlhCAANAPcAAAAAAAEBAQICAgMDAwQEBAUFBQYGBgcHBwgICAkJCQoKCgsLCwwMDA0NDQ4O
Dg8PDxAQEBERERISEhMTExQUFBUVFRYWFhcXFxgYGBkZGRoaGhsbGxwcHB0dHR4eHh8fHyAgICEh
ISIiIiMjIyQkJCUlJSYmJicnJygoKCkpKSoqKisrKywsLC0tLS4uLi8vLzAwMDExMTIyMjMzMzQ0
NDU1NTY2Njc3Nzg4ODk5OTo6Ojs7Ozw8PD09PT4+Pj8/P0BAQEFBQUJCQkNDQ0REREVFRUZGRkdH
R0hISElJSUpKSktLS0xMTE1NTU5OTk9PT1BQUFFRUVJSUlNTU1RUVFVVVVZWVldXV1hYWFlZWVpa
WltbW1xcXF1dXV5eXl9fX2BgYGFhYWJiYmNjY2RkZGVlZWZmZmdnZ2hoaGlpaWpqamtra2xsbG1t
bW5ubm9vb3BwcHFxcXJycnNzc3R0dHV1dXZ2dnd3d3h4eHl5eXp6ent7e3x8fH19fX5+fn9/f4CA
gIGBgYKCgoODg4SEhIWFhYaGhoeHh4iIiImJiYqKiouLi4yMjI2NjY6Ojo+Pj5CQkJGRkZKSkpOT
k5SUlJWVlZaWlpeXl5iYmJmZmZqampubm5ycnJ2dnZ6enp+fn6CgoKGhoaKioqOjo6SkpKWlpaam
pqenp6ioqKmpqaqqqqurq6ysrK2tra6urq+vr7CwsLGxsbKysrOzs7S0tLW1tba2tre3t7i4uLm5
ubq6uru7u7y8vL29vb6+vr+/v8DAwMHBwcLCwsPDw8TExMXFxcbGxsfHx8jIyMnJycrKysvLy8zM
zM3Nzc7Ozs/Pz9DQ0NHR0dLS0tPT09TU1NXV1dbW1tfX19jY2NnZ2dra2tvb29zc3N3d3d7e3t/f
3+Dg4OHh4eLi4uPj4+Tk5OXl5ebm5ufn5+jo6Onp6erq6uvr6+zs7O3t7e7u7u/v7/Dw8PHx8fLy
8vPz8/T09PX19fb29vf39/j4+Pn5+fr6+vv7+/z8/P39/f7+/v///ywAAAAACAANAAAILQABwGFG
sCAAZgMLMjuIUCHDhgQfQpTYkCKzfxT/LVSocSPBjh5BbhT5USHBgAA=
END_OF_DATA

my $mw = Tk::MainWindow->new(-title => 'TAOS/Parallax TCS230 "Tuna Can" Compass v1.0');
$mw->withdraw;

my $phoHSV = $mw->Photo(-data => $imgHSV, -format => 'gif');
my $phoLeft = $mw->Photo(-data => $imgLeft, -format => 'gif');
my $phoRight = $mw->Photo(-data => $imgRight, -format => 'gif');

$mw->fontCreate('mainfont', -family => 'Helvetica', -size => 100, -weight => 'bold');
my $FontScale = 96 / $mw->fontMeasure('mainfont', "M");
$mw->fontConfigure('mainfont', -size => int($FontScale * 12));

$mw->configure(-menu => my $mnuBar = $mw->Menu);
my $mnuFile = $mnuBar->cascade(-label => '~File');
$mnuFile->command(-label => '~Open Compass Colors', -command => \&OpenCompass, -font => 'mainfont');
$mnuFile->command(-label => '~Save Compass Colors', -command => \&SaveCompass, -font => 'mainfont');
$mnuFile->command(-label => 'Create ~BASIC Stamp Code', -command => \&CreateBasic, -font => 'mainfont');
$mnuFile->command(-label => 'Create New Color ~Disc', -command => \&CreateDisc, -font => 'mainfont');

my $fraTop = $mw->Frame->pack(-side => 'top');
my $fraSwatchPanel = $fraTop->Frame(
	-height => 256,
	-width => 128,
)->pack(-side => 'left');
my $fraRight = $fraTop->Frame->pack(-side => 'right');

my $fraBottom = $mw->Frame->pack(-side => 'bottom', -expand => 1, -fill => 'x');

my $Message;
my $txtMsg = $fraSwatchPanel->Label(
	-background => 'BLACK',
	-foreground => 'YELLOW',
	-textvariable => \$Message,
	-relief => 'sunken',
	-borderwidth => 2
)->pack(-side => 'top', -expand => 1, -fill => 'x', -padx => 2, -pady => 2);

$fraSwatchPanel->Label(-text => 'Red, Green & Blue', -font => 'mainfont')->pack;

my $canBars = $fraSwatchPanel->Canvas(
	-background => 'BLACK',
	-width => 128,
	-height => 60,
	-relief => 'sunken',
	-borderwidth => 2
)->pack(-side => 'top', -pady => 2);

my (@canBar, @txtBar);
foreach (0..2) {
	$canBar[$_] = $canBars->createRectangle(
		0, $_ * 20 + 4, -128, $_ * 20 + 22,
		-fill => $RGB[$_]
	);
	$txtBar[$_] = $canBars->createText(
		65,$_ * 20 + 13, 
		-text => '',
		-font => 'Helvetica 11 bold',
		-fill => 'WHITE',
		-justify => 'right'
	);
	my $btn = $canBars->Button(-image => $phoLeft, -command => eval("sub {ChangePeriod($_, -1)}"));
	$canBars->createWindow(11, $_ * 20 + 14, -window => $btn);
	$btn = $canBars->Button(-image => $phoRight, -command => eval("sub {ChangePeriod($_, 1)}"));
	$canBars->createWindow(125, $_ * 20 + 14, -window => $btn);
}

$fraSwatchPanel->Label(-text => 'Hue & Saturation', -font => 'mainfont')->pack;

my $canHSV = $fraSwatchPanel->Canvas(
	-background => 'BLACK',
	-height => 128,
	-width => 128,
	-relief => 'sunken',
	-borderwidth => 2
)->pack(-side => 'top', -padx => 2, -pady => 2);

$canHSV->createImage(67, 67, -image => $phoHSV);
my $ptrHSV = $canHSV->createOval(63, 63, 71, 71, -fill => 'YELLOW', -outline => 'BLACK');

my $CurrentGamma = 35;

my $fraCompass = $fraRight->Frame(
)->pack(-side => 'top');
my ($cSize, $cCenter) = (262, 133);
my $canCompass = $fraCompass->Canvas(
	-width => $cSize,
	-height => $cSize,
	-background => 'BLACK',
	-relief => 'sunken',
	-borderwidth => 2
)->pack(-padx => 2, -pady => 2);

my @ptrCompass;
my @diamond = (-30, 0, -15, -5, 0, 0, -15, 5);
foreach my $dir (0 .. 7) {
	my $th = $dir * 6.2831854 / 8;
	my ($st, $ct) = (sin($th), cos($th));
	my @coords = map {(
		int($cCenter + 126 * $ct + $ct * $diamond[$_] - $st * $diamond[$_ + 1]),
		int($cCenter + 126 * $st + $ct * $diamond[$_ + 1] + $st * $diamond[$_])
	)} (0, 2, 4, 6);
	$ptrCompass[$dir] = $canCompass->createPolygon(@coords, -fill => 'LIGHTGRAY', -outline => 'LIGHTGRAY');
	$canCompass->bind($ptrCompass[$dir], '<Button-1>', eval("sub{DefineCurrentHue($dir, 1)}"));
	$canCompass->bind($ptrCompass[$dir], '<Button-3>', eval("sub{DefineCurrentHue($dir, 0)}"));
}
my @Arrow = (-75, 0, -90, -15, -65, -15, -60, -10, 65, -10, 65, -15,
	90, 0, 65, 15, 65, 10, -60, 10, -65, 15, -90, 15);
my @coords = map {$_ + $cCenter} @Arrow;
my $ptrArrow = $canCompass->createPolygon(@coords, -fill => 'BLACK', -outline => 'BLACK', -width => 2);
my $dirSize = int(18 * $FontScale + 0.5);
my $LowDir = $canCompass->createText($cCenter, $cCenter - 79, -text => 'N', -font => "Helvetica $dirSize bold", -justify => 'center', -fill => 'YELLOW');
$canCompass->createText($cCenter, $cCenter + 79, -text => 'S', -font => "Helvetica $dirSize bold", -justify => 'center', -fill => 'YELLOW');
$canCompass->createText($cCenter - 79, $cCenter, -text => 'W', -font => "Helvetica $dirSize bold", -justify => 'center', -fill => 'YELLOW');
my $HighDir = $canCompass->createText($cCenter + 79, $cCenter, -text => 'E', -font => "Helvetica $dirSize bold", -justify => 'center', -fill => 'YELLOW');
my $ovlBearing = $canCompass->createOval(($cCenter - 44) x 2, ($cCenter + 44) x 2, -fill => 'BLACK', -outline => 'LIGHTGRAY', -width => 10);
$canCompass->createText(
	$cCenter, $cCenter - 9,
	-text => "BEARING",
	-font => 'Helvetica '.int(13 * $FontScale).' bold',
	-justify => 'center',
	-fill => 'GREEN'
);
my $txtBearing = $canCompass->createText(
	$cCenter, $cCenter + 16,
	-text => "---\xb0",
	-font => 'Helvetica '.int(24 * $FontScale).' bold',
	-justify => 'center',
	-fill => 'GREEN'
);

$fraBottom->Label(
	-text => "\xa9 Copyright 2003 Bueno Systems, Inc.",
	-font => 'Helvetica '.int(10 * $FontScale).' normal'
)->pack;

$mw->update;
$mw->deiconify;
$mw->raise;

my $Port;
foreach my $dev ((@Devices) x 2, undef) {
	unless (defined $dev) {
		$mw->messageBox(
			-title => 'Serial I/O Error',
			-message => 'Unable to locate a running TCS230 on any serial port.',
			-type => 'OK'
		);
		exit
	}
	$Message = "Scanning $dev.";
	$txtMsg->update;
	
	if ($Port = &$NewSerialPort($dev, 1)) {
		$Port->databits(8);
		$Port->baudrate(9600);
		$Port->parity("none");
		$Port->stopbits(2);
		$Port->handshake("none");
		$Port->buffers(4096,4096);
		$Port->write_settings;
	} else {
		next
	}
	&ResetStamp;
	sleep 1.0;
	my $stream = $Port->input;
	last if $stream =~ m/R\d\d\d G\d\d\d B\d\d\d\r/;
	$Port->close;
	undef $Port;
}

$Message = '';
my $UpdateID = $mw->repeat(50, \&Update);
MainLoop();

sub ChangePeriod {
	my ($index, $dir) = @_;
	&ResetStamp;
	$Periods[$index] += $dir;
	$Periods[$index] = 1 if $Periods[$index] < 1;
	$Periods[$index] = 255 if $Periods[$index] > 255;
	sleep .25;
	foreach (@Periods) {
		$Port->write(chr($_));
		sleep .025
	}
	sleep .5;
}

sub ResetStamp {
	undef @Periods;
	$Port->dtr_active(1);
	sleep .002;
	$Port->dtr_active(0);
	my $t0 = time;
	do {
		$stream .= $Port->input
	} until $stream =~ m/pR(\d\d\d) pG(\d\d\d) pB(\d\d\d)\r/osg || time - $t0 > .5;
	@Periods = map {$_ + 0} ($1, $2, $3) if defined $3;
	#print join(',', @Periods)."\n"
}

sub Update {
	my (@color, $nxt);
	$Port->error_msg(1);
	my $inp = $Port->input;
	return if $Busy;
	if ($inp) {
		$stream .= $inp;
		while ($stream =~ m/(R\d\d\d G\d\d\d B\d\d\d)\r/osg) {
			$lastRGB = $1;
			$nxt = pos($stream);
		}
		$stream = substr($stream, $nxt) if defined $nxt;
		if ($lastRGB) {
			foreach my $i (0..2) {
				$color[$i] = substr($lastRGB, $i * 5 + 1, 3) + 0;
			}
			@CurrentRGB = @color;
			my $status = grep {$_ > 255} @CurrentRGB;
			$Message = $status ? 'Saturated' : '';
			@CurrentRGB = map {$_ > 255 ? 255 : $_} @CurrentRGB;
			foreach my $i (0..2) {
				$canBars->itemconfigure($txtBar[$i], -text => $CurrentRGB[$i]);
				$canBars->coords($canBar[$i], 20, $i * 20 + 4, 20 + int($CurrentRGB[$i] * 0.375), $i * 20 + 22)
			}
			@CurrentHSV = RGB2HSV(@CurrentRGB);
			my $th = $CurrentHSV[0] * 6.2831854 / 256;
			my $r = $CurrentHSV[1] / 4;
			my ($ix, $iy) = (67 - int($r * cos($th)), 67 - int($r * sin($th)));
			$canHSV->coords($ptrHSV, $ix - 4, $iy - 4, $ix + 4, $iy + 4);
			$canCompass->itemconfigure($ovlBearing, -outline => ScreenColor(@CurrentRGB));
			if ($AllDefined) {
				my $th =  $HueMap[int($CurrentHSV[0])] * 6.2831854 / 256;
				my ($ct, $st) = (cos($th), sin($th));
				my @coords = map {(
					int($cCenter + $ct * $Arrow[2 * $_] - $st * $Arrow[2 * $_ + 1]),
					int($cCenter + $ct * $Arrow[2 * $_ + 1] + $st * $Arrow[2 * $_])
				)} (0 .. @Arrow / 2 - 1);
				$canCompass->coords($ptrArrow, @coords);
				$canCompass->itemconfigure(
					$txtBearing,
					-text => sprintf("%3.1d\xb0", int(($th * 360 / 6.2831854 + 90) / 5) * 5 % 360)
				)
			}
		}
	}
}

sub SetGamma {
	foreach (grep defined, @DefinedHue) {
		my ($n, @rgb) = @$_;
		#$lblColor[$n]->configure(-background => ScreenColor(@rgb))
	}
}

sub ScreenColor {
	return sprintf('#%2.2X%2.2X%2.2X', GammaCorrect(@_))
}

sub GammaCorrect {
	my $max = (sort {$b <=> $a} @_)[0];
	my $gammaf = ($max / 256) ** ($CurrentGamma / 100) * 256 / $max;
	return map {$gammaf * $_} @_
}

sub DefineCurrentHue {
	DefineHue(@_, $CurrentHSV[0])
}

sub DefineHue {
	return if $Busy;
	my ($dir, $def, $hue) = @_;
	my $color = $def ? ScreenColor(HSV2RGB($hue, 255, 255)) : 'LIGHTGRAY';
	$canCompass->itemconfigure($ptrCompass[$dir], -fill => $color);
	$DefinedHue[$dir] = $def ? $hue : undef;
	MakeHueMap() if $AllDefined = (grep {defined $_} @DefinedHue) == 8;
	if ($AllDefined) {
		$canCompass->itemconfigure($ptrArrow, -fill => 'RED', -outline => 'WHITE');
		$canCompass->raise($ptrArrow, $HighDir)
	} else {
		$canCompass->itemconfigure($ptrArrow, -fill => 'BLACK', -outline => 'BLACK');
		$canCompass->lower($ptrArrow, $LowDir);
		$canCompass->itemconfigure($txtBearing, -text => "---\xb0")
	}
}

sub RGB2HSV {
	my ($red, $grn, $blu) = map {$_ / 256} @_;
	my ($max, $min) = (sort {$a <=> $b} ($red, $grn, $blu))[2, 0];
	return (0, 0, $max * 256) if $max == $min;
	my $val = $max * 256;
	my $sat = int(($max - $min) * 256 / $max);
	my $hue;
	if ($red == $max) {
		$hue = ($grn - $blu) / ($max - $min)
	} elsif ($grn == $max) {
		$hue = 2 + ($blu - $red) / ($max - $min)
	} else {
		$hue = 4 + ($red - $grn) / ($max - $min)
	}
	$hue /= 6;
	$hue += 1 if $hue < 0;
	$hue *= 256;
	return ($hue, $sat, $val)
}

sub HSV2RGB {
	my ($hue, $sat, $val) = map {$_ / 256} @_;
	my @rgb;
	return (0, 0, 0) if $val == 0;
	return ($val * 256) x 3 if $sat == 0;
	my $i = int($hue * 6);
	my $f = $hue * 6 - $i;
	my $t1 = $val * (1 - $sat) * 256;
	my $t2 = $val * (1 - $sat * $f) * 256;
	my $t3 = $val * (1 - $sat * (1 - $f)) * 256;
	my $v = $val * 256;
	return @{([$v, $t3, $t1], [$t2, $v, $t1], [$t1, $v, $t3], [$t1, $t2, $v], [$t3, $t1, $v], [$v, $t1, $t2])[$i]}
}

sub between {
	my ($x0, $x, $x1) = @_;
	if (abs($x - $x1) > 128) {
		if ($x < $x1) {
			$x1 += 256
		} else {
			$x += 256
		}
	} elsif (abs($x - $x0) > 128) {
		if ($x < $x0) {
			$x0 += 256
		} else {
			$x += 256
		}
	}
	return $x0 <= $x && $x <= $x1 || $x0 >= $x && $x >= $x1;
}

sub OpenCompass {
	return if $Busy;
	my $filename = $mw->getOpenFile(
		-title => 'Open Compass Color File',
		-filetypes => [['Compass Files', '.cps']],
		-defaultextension => '.cps',
		-initialdir => $DefaultDir
	);
	return unless defined $filename && $filename ne '';
	($DefaultDir = (fileparse($filename))[1]) =~ s:/:$DirDelim:g;
	unless ($filename =~ m/\.cps$/) {
		$mw->messageBox(
			-title => 'Open Compass Color File',
			-message => 'File name must end in ".cps".',
			-type => 'OK'
		);
		return
	}
	unless (open CPS, "<$filename") {
		$mw->messageBox(
			-title => 'Open Compass Color File',
			-message => "Error reading from $filename: $!",
			-type => 'OK'
		);
		return
	}
	my $txt = <CPS>;
	unless ($txt =~ m/^#TCS230 Compass Color Data\r?\n((\d+,){10}(\d+))\r?\n$/s) {
		$mw->messageBox(
			-title => 'Open Compass Color File',
			-message => "Corrupt color data in $filename.",
			-type => 'OK'
		);
		return
	}
	my @data = split(/,/, $1);
	@Periods = @data[0 .. 2];
	ChangePeriod(0, 0);
	@DefinedHue = undef;
	foreach (0 .. 7) {
		DefineHue($_, 1, $data[$_ + 3])
	}
}	
	
sub SaveCompass {
	return if $Busy;
	unless ($AllDefined) {
		$mw->messageBox(
			-title => 'Save Compass Colors',
			-message => 'Unable to save compass colors unless all eight compass points are defined.',
			-type => 'OK'
		);
		return
	}
	my $filename = $mw->getSaveFile(
		-title => 'Save Compass Colors As ...',
		-filetypes => [['Compass Files', '.cps']],
		-defaultextension => '.cps',
		-initialdir => $DefaultDir
	);
	return unless defined $filename && $filename ne '';
	($DefaultDir = (fileparse($filename))[1]) =~ s:/:$DirDelim:g;
	unless ($filename =~ m/\.cps$/) {
		$mw->messageBox(
			-title => 'Save Compass Colors',
			-message => 'File name must end in ".cps".',
			-type => 'OK'
		);
		return
	}
	unless (open CMP, ">$filename") {
		$mw->messageBox(
			-title => 'Save Compass Colors',
			-message => "Error writing to $filename: $!",
			-type => 'OK'
		);
		return
	}	
	print CMP "#TCS230 Compass Color Data\n";
	print CMP join (',', map {int($_)} (@Periods, @DefinedHue)), "\n";
	close CMP
}

sub CreateBasic {
	return if $Busy;
	unless ($AllDefined) {
		$mw->messageBox(
			-title => 'Save BASIC Stamp Program',
			-message => 'Unable to BASIC Stamp Program unless all eight compass points are defined.',
			-type => 'OK'
		);
		return
	}
	my $filename = $mw->getSaveFile(
		-title => 'Save BASIC Stamp Program As ...',
		-filetypes => [['BASIC Stamp 2 Source', '.bs2']],
		-defaultextension => '.bs2',
		-initialdir => $DefaultDir
	);
	return unless defined $filename && $filename ne '';
	($DefaultDir = (fileparse($filename))[1]) =~ s:/:$DirDelim:g;
	unless ($filename =~ m/\.bs2$/) {
		$mw->messageBox(
			-title => 'Save BASIC Stamp Program',
			-message => 'File name must end in ".bs2".',
			-type => 'OK'
		);
		return
	}
	unless (open BS2, ">$filename") {
		$mw->messageBox(
			-title => 'Save BASIC Stamp Program',
			-message => "Error writing to $filename: $!",
			-type => 'OK'
		);
		return
	}	
	print BS2 <<END_OF_BS2;
'{\$STAMP BS2}
'{\$PBASIC 2.5}
 
'Define port pins

EN	con	1
A0	con	2
S0	con	3
S1	con	4
S2	con	5
S3	con	6
nLED	con	7
OUT	con	8

'Define variables for color results.

pRED	var	byte
pGREEN	var	byte
pBLUE	var	byte

RED	var	word
GREEN	var	word
BLUE	var	word
HUE	var	byte
cmax	var	byte
cmin	var	byte
DIR	var	word

'Program starts here.

Start:	low	A0			'For Unit 0; use high for Unit 1.
	high	S0			'Maximum output rate.
	high	S1			'        "
	low	nLED			'Turn on LED.
	gosub	GetPeriods		'Get count periods for RGB.
	
'-----------------------YOUR CODE GOES BELOW THIS LINE-----------------------	

MainLp:	gosub	GetDir			'Main loop
	debug	dec3 DIR, cr		'SAMPLE CODE: Read & echo direction.
	goto	MainLp			'Back for another.

'-----------------------YOUR CODE GOES ABOVE THIS LINE-----------------------	

'GetDir: Read the current direction of travel.

GetDir:	gosub	Color			'Read the compass color.
	gosub	RGBtoHue		'Convert RGB to hue.
	read	HUE + 3, DIR		'Use hue to look up direction.
	return				'  and return.

'Color: Read all three color components.

Color:	high	EN			'Enable output/turn on LEDs.
	low	S2		   	'Address the red output.
	low	S3
	count	OUT, pRED, RED		'Read the red component.
	high	S3			'Address the blue output.
	count	OUT, pBLUE, BLUE	'Read the blue component.
	high	S2			'Address the green output.
	count	OUT, pGREEN, GREEN	'Read the green component.
	low	EN			'Disable output/turn off LEDs.
	return				'Return.

'RGBToHue: Extract Hue from RGB components.

RGBToHue:
	RED = RED max 255
	GREEN = GREEN max 255
	BLUE = BLUE max 255		
	cmax = RED min GREEN min BLUE	'Find maximum RGB component.
	cmin = RED max GREEN max BLUE	'Find minimum RGB component.
	if cmin = cmax then		'Hue is undefined if SAT is zero.
	  HUE = 0
	  return
	endif
	if (cmax = RED) then		'Find hue in color circle.
	  HUE = abs(GREEN - BLUE) * 43 / (cmax - cmin)
	  if BLUE > GREEN then HUE = 256 - HUE
	else 
	  if (cmax = GREEN) then
	    HUE = abs(BLUE - RED) * 43 / (cmax - cmin)
	    if RED > BLUE then HUE = 256 - HUE
            HUE = HUE + 85
	  else
	    HUE = abs(RED - GREEN) * 43 / (cmax - cmin)
            if GREEN > RED then HUE = 256 - HUE
	    HUE = HUE + 170
	  endif
	endif
	return				'Return.

'GetPeriods: Read count periods for RGB from EEPROM.

GetPeriods:
	read	0, pRED
	read	1, pGREEN
	read	2, pBLUE
	return

'Data: Count periods for RGB components.

Periods:
	data $Periods[0], $Periods[1], $Periods[2]

'Data: Lookup table for converting Hue to Direction.

Directions:
END_OF_BS2

	for (my $line = 0; $line < 255; $line += 8) {
		print BS2 "\tdata\t";
		print BS2 join(', ', map {sprintf('%3.1d', int(($HueMap[$_] + 64) % 256))} ($line .. $line + 7)), "\n"		
	}
	print BS2 "\tend\n";
	close BS2
}	

sub CreateDisc {
	return if $Busy;
	unless ($AllDefined) {
		$mw->messageBox(
			-title => 'Create Color Disc',
			-message => 'Unable to create a new color disc until all eight compass points are defined.',
			-type => 'OK'
		);
		return
	}
	my $types = [
		['Jpeg Files', '.jpg'],
	];
	my $filename = $mw->getSaveFile(
		-filetypes=>$types,
		-title => 'Save Color Disc Image As ...',
		-defaultextension => '.jpg',
		-initialdir => $DefaultDir
	);
	return unless defined $filename && $filename ne '';
	($DefaultDir = (fileparse($filename))[1]) =~ s:/:$DirDelim:g;
	unless ($filename =~ m/\.jpg$/) {
		$mw->messageBox(
			-title => 'Create Color Disc',
			-message => 'File name must end in ".jpg".',
			-type => 'OK'
		);
		return
	}
	my $img = GD::Image->new(256, 256);
	my @palette = map {$img->colorAllocate($_ ? HSV2RGB($_, 255, 255) : (255, 255, 255))} (0 .. 255);
	$Busy = 1;
	my ($hm0, $hm63) = @HueMap[0, 63];
	my $rev = $hm0 > $hm63 && abs($hm0 - $hm63) < 128 || $hm0 < $hm63 && abs($hm0 - $hm63) >= 128;
	my @CorrectMap = map {(256 + ($rev ? $hm0 - $_ : $_ - $hm0)) % 256} @HueMap;
	print join('|', @HueMap)."\n";
	print join('|', @CorrectMap)."\n";
	foreach my $i (0 .. 255) {
		$Message = "Creating file: " . (255 - $i);
		$txtMsg->update;
		foreach my $j (0 .. 255) {
			my $th = atan2($j - 128, $i - 128) + 3.14159265;
			my $hueindex = int($th * 255 / 6.2831854 + .999);
			my $d = sqrt(($i - 128) ** 2 + ($j - 128) ** 2);
			if ($d >= 10 && $d <128) {
				$img->setPixel($i, $j, $palette[$CorrectMap[$hueindex] || 1])
			}
		}
	}
	$Busy = 0;
	if (open(JPG, ">$filename")) {
		binmode JPG;
		print JPG $img->jpeg(100);
		close JPG
	} else {
		$mw->messageBox(
			-title => 'Create Color Disc',
			-message => "Error writing to $filename: $!",
			-type => 'OK'
		);
		return
	}	
}

sub MakeHueMap {
	my @indices = sort {$DefinedHue[$a] <=> $DefinedHue[$b]} (0 .. @DefinedHue - 1);
	foreach (0 .. @indices - 1) {
		my $dif = abs($indices[$_] - $indices[($_ + 1) % 8]);
		unless ($dif == 1 || $dif == 7) {
			$mw->messageBox(
				-title => 'Define Compass Points',
				-message => 'The eight compass points need to be in order by hue. Please redefine them.',
				-type => 'OK'
			);
			$AllDefined = 0;
			return
		}
	}
	my $index = 7;
	my $currenthue = $DefinedHue[$indices[$index]] - 256;
	my $nexthue = $DefinedHue[$indices[($index + 1) % 8]];
	foreach my $hue (0 .. 255) {
		if ($hue >= $nexthue) {
			$index = ($index + 1) % 8;
			$currenthue = $nexthue;
			$nexthue = $DefinedHue[$indices[($index + 1) % 8]];
			$nexthue += 256 if $nexthue < $currenthue
		}
		my $p = ($hue - $currenthue) / ($nexthue - $currenthue);
		my $h0 = $indices[$index];
		my $h1 = $indices[($index + 1) % 8];
		if (abs($h1 - $h0) > 1) {
			if ($h1 > $h0) {
				$h0 += 8
			} else {
				$h1 += 8
			}
		}
		my $value = int(($p * $h1 + (1 - $p) * $h0) * 32);
		$value += 256 if $value < 0;
		$value -= 256 if $value > 255;
		$HueMap[$hue] = $value;
	}
}
		
sub GetColorWheel {
	return <<END_OF_DATA;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END_OF_DATA
}